==> wsclean_2.2.1.orig.tar.bz2: wsclean-2.2.1/deconvolution/genericclean.cpp <==
#include "genericclean.h"
#include "../multiscale/clarkloop.h"

#include "../lane.h"

#include "../multiscale/threadeddeconvolutiontools.h"

#include <boost/thread/thread.hpp>

GenericClean::GenericClean(ImageBufferAllocator& allocator, bool clarkOptimization) :
	_convolutionPadding(1.1),
	_useClarkOptimization(clarkOptimization),
	_allocator(allocator)
{
}

void GenericClean::ExecuteMajorIteration(ImageSet& dirtySet, ImageSet& modelSet, const ao::uvector<const double*>& psfs, size_t width, size_t height, bool& reachedMajorThreshold)
{
	Logger::Debug << "Running dynamically-joined clean algorithm.\n";
	
	const size_t iterationCounterAtStart = _iterationNumber;
	if(_stopOnNegativeComponent)
		_allowNegativeComponents = true;
	_width = width;
	_height = height;
	_convolutionWidth = ceil(_convolutionPadding * _width);
	_convolutionHeight = ceil(_convolutionPadding * _height);
	if(_convolutionWidth%2 != 0)
		++_convolutionWidth;
	if(_convolutionHeight%2 != 0)
		++_convolutionHeight;
	
	ImageBufferAllocator::Ptr integrated, scratchA, scratchB;
	_allocator.Allocate(width*height, integrated);
	_allocator.Allocate(_convolutionWidth*_convolutionHeight, scratchA);
	_allocator.Allocate(_convolutionWidth*_convolutionHeight, scratchB);
	dirtySet.GetSquareIntegrated(integrated.data(), scratchA.data());
	size_t componentX=0, componentY=0;
	double maxValue = findPeak(integrated.data(), componentX, componentY);
	Logger::Info << "Initial peak: " << peakDescription(integrated.data(), componentX, componentY) << '\n';
	double firstThreshold = this->_threshold;
	double stopGainThreshold = std::fabs(maxValue)*(1.0-this->_mGain);
	if(stopGainThreshold > firstThreshold)
	{
		firstThreshold = stopGainThreshold;
		Logger::Info << "Next major iteration at: " << stopGainThreshold << '\n';
	}
	else if(this->_mGain != 1.0) {
		Logger::Info << "Major iteration threshold reached global threshold of " << this->_threshold << ": final major iteration.\n";
	}
		
	if(_useClarkOptimization)
	{
		size_t startIteration = _iterationNumber;
		ClarkLoop clarkLoop(_width, _height, _convolutionWidth, _convolutionHeight);
		clarkLoop.SetIterationInfo(_iterationNumber, MaxNIter());
		clarkLoop.SetThreshold(firstThreshold);
		clarkLoop.SetGain(Gain());
		clarkLoop.SetAllowNegativeComponents(AllowNegativeComponents());
		clarkLoop.SetSpectralFitter(&Fitter());
		if(_cleanMask)
			clarkLoop.SetMask(_cleanMask);
		const size_t
			horBorderSize = round(_width * CleanBorderRatio()),
			vertBorderSize = round(_height * CleanBorderRatio());
		clarkLoop.SetCleanBorders(horBorderSize, vertBorderSize);
		
		maxValue = clarkLoop.Run(dirtySet, psfs);
		
		_iterationNumber = clarkLoop.CurrentIteration();
		
		Logger::Info << "Performed " << (_iterationNumber - startIteration) << " / " << _iterationNumber << " iterations with Clark optimization in this major iteration.\n";
		
		for(size_t imageIndex=0; imageIndex!=dirtySet.size(); ++imageIndex)
		{
			// TODO this can be multi-threaded if each thread has its own temporaries
			const double *psf = psfs[dirtySet.PSFIndex(imageIndex)];
			clarkLoop.CorrectResidualDirty(scratchA.data(), scratchB.data(), integrated.data(), imageIndex, dirtySet[imageIndex],  psf);
			
			clarkLoop.GetFullIndividualModel(imageIndex, scratchA.data());
			double* model = modelSet[imageIndex];
			for(size_t i=0; i!=_width*_height; ++i)
				model[i] += scratchA.data()[i];
		}
	}
	else {
		ThreadedDeconvolutionTools tools(_threadCount);
		size_t peakIndex = componentX + componentY*_width;

		ao::uvector<double> peakValues(dirtySet.size());
		
		while(fabs(maxValue) > firstThreshold && this->_iterationNumber < this->_maxIter && !(maxValue<0.0 && this->_stopOnNegativeComponent))
		{
			if(this->_iterationNumber <= 10 ||
				(this->_iterationNumber <= 100 && this->_iterationNumber % 10 == 0) ||
				(this->_iterationNumber <= 1000 && this->_iterationNumber % 100 == 0) ||
				this->_iterationNumber % 1000 == 0)
				Logger::Info << "Iteration " << this->_iterationNumber << ": " << peakDescription(integrated.data(), componentX, componentY) << '\n';
			
			for(size_t i=0; i!=dirtySet.size(); ++i)
				peakValues[i] = dirtySet[i][peakIndex];
			
			PerformSpectralFit(peakValues.data());
			
			for(size_t i=0; i!=dirtySet.size(); ++i)
			{
				peakValues[i] *= this->_gain;
				modelSet[i][peakIndex] += peakValues[i];
				
				size_t psfIndex = dirtySet.PSFIndex(i);
				
				tools.SubtractImage(dirtySet[i], psfs[psfIndex], width, height, componentX, componentY, peakValues[i]);
			}
			
			dirtySet.GetSquareIntegrated(integrated.data(), scratchA.data());
			maxValue = findPeak(integrated.data(), componentX, componentY);
			
			peakIndex = componentX + componentY*_width;
			
			++this->_iterationNumber;
		}
	}
	Logger::Info << "Stopped on peak " << maxValue << '\n';
	reachedMajorThreshold = std::fabs(maxValue) <= stopGainThreshold && (maxValue != 0.0) && (_iterationNumber-iterationCounterAtStart)!=0;
}

std::string GenericClean::peakDescription(const double* image, size_t& x, size_t& y)
{
	std::ostringstream str;
	size_t index = x + y*_width;
	double peak = image[index];
	str << peak << " Jy at " << x << "," << y;
	return str.str();
}

double GenericClean::findPeak(const double* image, size_t& x, size_t& y)
{
	if(_cleanMask == 0)
		return SimpleClean::FindPeak(image, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanBorderRatio);
	else
		return SimpleClean::FindPeakWithMask(image, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanMask, _cleanBorderRatio);
}

==> wsclean_2.4.orig.tar.bz2: wsclean-2.4/deconvolution/genericclean.cpp <==
#include "genericclean.h"

#include "clarkloop.h"

#include "../lane.h"

#include "../multiscale/threadeddeconvolutiontools.h"

#include "../units/fluxdensity.h"

#include <boost/thread/thread.hpp>

GenericClean::GenericClean(ImageBufferAllocator& allocator, bool clarkOptimization) :
	_convolutionPadding(1.1),
	_useClarkOptimization(clarkOptimization),
	_allocator(allocator)
{
}

void GenericClean::ExecuteMajorIteration(ImageSet& dirtySet, ImageSet& modelSet, const ao::uvector<const double*>& psfs, size_t width, size_t height, bool& reachedMajorThreshold)
{
	const size_t iterationCounterAtStart = _iterationNumber;
	if(_stopOnNegativeComponent)
		_allowNegativeComponents = true;
	_width = width;
	_height = height;
	_convolutionWidth = ceil(_convolutionPadding * _width);
	_convolutionHeight = ceil(_convolutionPadding * _height);
	if(_convolutionWidth%2 != 0)
		++_convolutionWidth;
	if(_convolutionHeight%2 != 0)
		++_convolutionHeight;
	
	ImageBufferAllocator::Ptr integrated, scratchA, scratchB;
	_allocator.Allocate(width*height, integrated);
	_allocator.Allocate(_convolutionWidth*_convolutionHeight, scratchA);
	_allocator.Allocate(_convolutionWidth*_convolutionHeight, scratchB);
	dirtySet.GetLinearIntegrated(integrated.data());
	size_t componentX=0, componentY=0;
	double maxValue = findPeak(integrated.data(), scratchA.data(), componentX, componentY);
	Logger::Info << "Initial peak: " << peakDescription(integrated.data(), componentX, componentY) << '\n';
	double firstThreshold = this->_threshold;
	double stopGainThreshold = std::fabs(maxValue)*(1.0-this->_mGain);
	if(stopGainThreshold > firstThreshold)
	{
		firstThreshold = stopGainThreshold;
		Logger::Info << "Next major iteration at: " << FluxDensity::ToNiceString(stopGainThreshold) << '\n';
	}
	else if(this->_mGain != 1.0) {
		Logger::Info << "Major iteration threshold reached global threshold of " << FluxDensity::ToNiceString(this->_threshold) << ": final major iteration.\n";
	}
	
	if(_useClarkOptimization)
	{
		size_t startIteration = _iterationNumber;
		ClarkLoop clarkLoop(_width, _height, _convolutionWidth, _convolutionHeight);
		clarkLoop.SetIterationInfo(_iterationNumber, MaxNIter());
		clarkLoop.SetThreshold(firstThreshold, firstThreshold*0.99);
		clarkLoop.SetGain(Gain());
		clarkLoop.SetAllowNegativeComponents(AllowNegativeComponents());
		clarkLoop.SetStopOnNegativeComponent(StopOnNegativeComponents());
		clarkLoop.SetSpectralFitter(&Fitter());
		if(!_rmsFactorImage.empty())
			clarkLoop.SetRMSFactorImage(_rmsFactorImage);
		if(_cleanMask)
			clarkLoop.SetMask(_cleanMask);
		const size_t
			horBorderSize = round(_width * CleanBorderRatio()),
			vertBorderSize = round(_height * CleanBorderRatio());
		clarkLoop.SetCleanBorders(horBorderSize, vertBorderSize);
		
		maxValue = clarkLoop.Run(dirtySet, psfs);
		
		_iterationNumber = clarkLoop.CurrentIteration();
		
		Logger::Info << "Performed " << (_iterationNumber - startIteration) << " / " << _iterationNumber << " iterations with Clark optimization in this major iteration.\n";
		
		for(size_t imageIndex=0; imageIndex!=dirtySet.size(); ++imageIndex)
		{
			// TODO this can be multi-threaded if each thread has its own temporaries
			const double *psf = psfs[dirtySet.PSFIndex(imageIndex)];
			clarkLoop.CorrectResidualDirty(scratchA.data(), scratchB.data(), integrated.data(), imageIndex, dirtySet[imageIndex],  psf);
			
			clarkLoop.GetFullIndividualModel(imageIndex, scratchA.data());
			double* model = modelSet[imageIndex];
			for(size_t i=0; i!=_width*_height; ++i)
				model[i] += scratchA.data()[i];
		}
	}
	else {
		ThreadedDeconvolutionTools tools(_threadCount);
		size_t peakIndex = componentX + componentY*_width;

		ao::uvector<double> peakValues(dirtySet.size());
		
		while(fabs(maxValue) > firstThreshold && this->_iterationNumber < this->_maxIter && !(maxValue<0.0 && this->_stopOnNegativeComponent))
		{
			if(this->_iterationNumber <= 10 ||
				(this->_iterationNumber <= 100 && this->_iterationNumber % 10 == 0) ||
				(this->_iterationNumber <= 1000 && this->_iterationNumber % 100 == 0) ||
				this->_iterationNumber % 1000 == 0)
				Logger::Info << "Iteration " << this->_iterationNumber << ": " << peakDescription(integrated.data(), componentX, componentY) << '\n';
			
			for(size_t i=0; i!=dirtySet.size(); ++i)
				peakValues[i] = dirtySet[i][peakIndex];
			
			PerformSpectralFit(peakValues.data());
			
			for(size_t i=0; i!=dirtySet.size(); ++i)
			{
				peakValues[i] *= this->_gain;
				modelSet[i][peakIndex] += peakValues[i];
				
				size_t psfIndex = dirtySet.PSFIndex(i);
				
				tools.SubtractImage(dirtySet[i], psfs[psfIndex], width, height, componentX, componentY, peakValues[i]);
			}
			
			dirtySet.GetSquareIntegrated(integrated.data(), scratchA.data());
			maxValue = findPeak(integrated.data(), scratchA.data(), componentX, componentY);
			
			peakIndex = componentX + componentY*_width;
			
			++this->_iterationNumber;
		}
	}
	Logger::Info << "Stopped on peak " << FluxDensity::ToNiceString(maxValue) << ", because ";
	bool
		maxIterReached = _iterationNumber >= MaxNIter(),
		finalThresholdReached = std::fabs(maxValue) <= _threshold || maxValue == 0.0,
		negativeReached = maxValue<0.0 && this->_stopOnNegativeComponent,
		mgainReached = std::fabs(maxValue) <= stopGainThreshold,
		didWork = (_iterationNumber-iterationCounterAtStart)!=0;
	
	if(maxIterReached)
		Logger::Info << "maximum number of iterations was reached.\n";
	else if(finalThresholdReached)
		Logger::Info << "the threshold was reached.\n";
	else if(negativeReached)
		Logger::Info << "a negative component was found.\n";
	else if(!didWork)
		Logger::Info << "no iterations could be performed.\n";
	else
		Logger::Info << "the minor-loop threshold was reached. Continuing cleaning after inversion/prediction round.\n";
	reachedMajorThreshold = mgainReached && didWork && !negativeReached && !finalThresholdReached;
}

std::string GenericClean::peakDescription(const double* image, size_t& x, size_t& y)
{
	std::ostringstream str;
	size_t index = x + y*_width;
	double peak = image[index];
	str << FluxDensity::ToNiceString(peak) << " at " << x << "," << y;
	return str.str();
}

double GenericClean::findPeak(const double* image, double* scratch, size_t& x, size_t& y)
{
	if(_rmsFactorImage.empty())
	{
		if(_cleanMask == 0)
			return SimpleClean::FindPeak(image, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanBorderRatio);
		else
			return SimpleClean::FindPeakWithMask(image, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanMask, _cleanBorderRatio);
	}
	else {
		for(size_t i=0; i!=_width*_height; ++i)
		{
			scratch[i] = image[i] * _rmsFactorImage[i];
		}
		double maxValue;
		if(_cleanMask == 0)
			maxValue = SimpleClean::FindPeak(scratch, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanBorderRatio);
		else
			maxValue = SimpleClean::FindPeakWithMask(scratch, _width, _height, x, y, _allowNegativeComponents, 0, _height, _cleanMask, _cleanBorderRatio);
		return maxValue;
	}
}
